
#include "ddxutils.h"
#include "ddclipres.h"

//
// Uses the DX7 interface
//

int ClearFrame(LPDIRECTDRAWSURFACE7 pBackBuffer, DWORD dwColor);
void DestroyDDObjects(void);

typedef struct _MYRGNDATA
{
	RGNDATAHEADER rdh;
	RECT rc[10];
}MYRGNDATA;

typedef struct _SPRITE                   // For the sprites
{
	int X;    // xpos
	int Y;    // ypos
	int W;    // width
	int H;    // height
	float dX; // velocityX
	float dY; // velocityY
	float fX; // these two fX & fY accumulate
	float fY; // the additions for position
}SPRITE;

COLORREF backCol;
MYRGNDATA  rgn;     // Includes the RGNDATAHEADER Type
HINSTANCE hInst;	// Instance handle
HWND hwndMain;		// Main window handle


// Define Consts for this app
#define NUMSPRITES     6
#define WINDOW_WIDTH   600
#define WINDOW_HEIGHT  420
#define SPRITE_WIDTH   84
#define SPRITE_HEIGHT  80

LPDIRECTDRAW7 		 lpDD; 			// DirectDraw7 object
LPDIRECTDRAWSURFACE7 lpDDSPrimary; 	// DirectDraw primary surface
LPDIRECTDRAWSURFACE7 lpDDSBack;		// DirectDraw back surface
LPDIRECTDRAWCLIPPER  lpDDClipBack;    // DirectDraw Clipper for back surface
LPDIRECTDRAWCLIPPER  lpDDClipPrimary; // DirectDraw Clipper for primary surface
LPDIRECTDRAWSURFACE7 lpBmpSurface;    // DirectDraw surface to hold the sprite

BOOL  bActive = TRUE; // Is App active
DWORD ddRet;          // Return value from calls to Direct Draw

DDSURFACEDESC2  ddsd;		// Surface Description Struct
DDSCAPS2		ddscaps;    // Capabilities of surface Struct
RECT			bkRect;

// the sprites
SPRITE Sprite[NUMSPRITES];

// Two separate timings are taken in function UpdateFrame(),
// one for the animation and one to display f.p.s.
int newTime;
int oldTime;

char * fName = "dx1.bmp"; // sprite image

LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

void DDInitError(char * s, DWORD code )
{
	MessageBox(NULL, DDXErrorString(code), "DirectX Error", MB_OK | MB_ICONEXCLAMATION);
}

int RestoreAll(void)
{
	int ddRet;

	ddRet = IDirectDrawSurface7_Restore(lpDDSPrimary);
	if (ddRet == DD_OK){
		ddRet = IDirectDrawSurface7_Restore(lpBmpSurface);
		ddRet = IDirectDrawSurface7_Restore(lpDDSBack);
	}
	return ddRet;
}

int Random(int x)
{
	return (rand()%x);
}

int FlipWindowed(HWND hwnd, LPDIRECTDRAWSURFACE7 pDDSPrimary, LPDIRECTDRAWSURFACE7 pDDSBack)
{
	RECT rcRectSrc;
	RECT rcRectDest;
	POINT pt;
	int ddret;

	if (pDDSPrimary == 0 || pDDSBack == 0)
		return E_POINTER;

	if (IsIconic(hwnd))
   		return 0;  // Important

	// First we need to figure out where the rectangle is on the screen
	pt.x = 0;
	pt.y = 0;
	ClientToScreen(hwnd, &pt);                     // Pos of the client
	GetClientRect(hwnd, &rcRectDest);              // Size of client
	OffsetRect(&rcRectDest, pt.x, pt.y);           // Offset it

	// IDirectDrawSurface7_Blt() performs arbitrary stretching between
	// the srouce and destination surfaces, if the size is
	// there is no stretching. Null for the Src RECT uses whole surface.
	ddret = IDirectDrawSurface7_Blt(pDDSPrimary, &rcRectDest, lpDDSBack, NULL, DDBLT_WAIT, NULL);

	if (ddret == DDERR_SURFACELOST)
		ddret = RestoreAll();

	return ddret;

}

void UpdateSprite(SPRITE * pSprite, float fTimeDelta)
{
	// Update the sprite position
	pSprite->fX += pSprite->dX * fTimeDelta;
	pSprite->fY += pSprite->dY * fTimeDelta;

	// Clip the position, and bounce if it hits the edge
	if (pSprite->fX < 0.0f){
		pSprite->fX  = 0.0f;
		pSprite->dX = -pSprite->dX;
	}

	if (pSprite->fX >= WINDOW_WIDTH - SPRITE_WIDTH){
		pSprite->fX = WINDOW_WIDTH - 1 - SPRITE_WIDTH;
		pSprite->dX = -(Random(15) + 10);
	}

	if (pSprite->fY < 0.0f){
		pSprite->fY = 0.0f;
		pSprite->dY = -pSprite->dY;
	}

	if (pSprite->fY > WINDOW_HEIGHT - SPRITE_HEIGHT){
		pSprite->fY = WINDOW_HEIGHT - 1 - SPRITE_HEIGHT;
		pSprite->dY = -(Random(15) + 10);
	}

	pSprite->X = pSprite->fX;
	pSprite->Y = pSprite->fY;
	pSprite->W = pSprite->X + SPRITE_WIDTH;
	pSprite->H = pSprite->Y + SPRITE_HEIGHT;
}

int CreateWindowedDisplay(HWND hwnd, DWORD dwWidth, DWORD dwHeight)
{
	ddRet = DirectDrawCreateEx(NULL, &lpDD, &IID_IDirectDraw7, NULL);
	if(ddRet != DD_OK){
		DDInitError("DirectDrawCreateEx Failed", ddRet);
		return 1;
	}

	ddRet = IDirectDraw7_SetCooperativeLevel(lpDD, hwndMain, DDSCL_NORMAL);
	if ( ddRet != DD_OK ){
		DDInitError("SetCooperativeLevel Failed", ddRet);
		return 1;
	}

	// Create the primary surface
	memset(&ddsd, 0, sizeof(ddsd));       // Clear structure
	ddsd.dwSize            = sizeof(ddsd);
	ddsd.dwFlags           = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE;

	ddRet = IDirectDraw7_CreateSurface(lpDD, &ddsd, &lpDDSPrimary, NULL);
	if ( ddRet != DD_OK ){
	    DDInitError("CreateSurface Failed", ddRet);
		return 1;
	}

	memset(&ddsd, 0, sizeof(ddsd));       // Clear structure
	ddsd.dwSize = sizeof(ddsd);
	ddRet = IDirectDrawSurface7_GetSurfaceDesc(lpDDSPrimary, &ddsd);
	if (ddRet != DD_OK){
		DDInitError("GetSurfaceDesc Failed", ddRet);
		return 1;
	}

	// We must find out if the display is 16 or 32 bit
	// and if it's RGB555 or RGB565 for 16 bit so as to
	// set the colour correctly.
	int r = 142, g = 102, b = 122;  // Our colour

	if (ddsd.ddpfPixelFormat.dwGBitMask      == 0x07E0)  // RGB565 (16bit)
		backCol = DDXRGB565(r, g, b);
	else if (ddsd.ddpfPixelFormat.dwGBitMask == 0x03E0)  // RGB555 (15bit)
		backCol = DDXRGB555(r, g, b);
	else if (ddsd.ddpfPixelFormat.dwGBitMask == 0xFF00)  // RGB888 (24/32bit)
		backCol = DDXRGB888(r, g, b);
	else
		backCol = 0; // For an 8 bit display the colours will not be set correctly

	// Create the backbuffer separately for a windowed display
	memset(&ddsd, 0, sizeof(ddsd));        // Clear structure
	ddsd.dwSize          = sizeof(ddsd);
	ddsd.dwFlags         = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps  = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth         = WINDOW_WIDTH;
	ddsd.dwHeight        = WINDOW_HEIGHT;

	ddRet = IDirectDraw7_CreateSurface(lpDD, &ddsd, &lpDDSBack, NULL);
	if ( ddRet != DD_OK ){
		DDInitError("CreateSurface Failed", ddRet);
		return 1;
	}

	// Setting up a Clip List is worth the effort
	rgn.rdh.dwSize   = sizeof(RGNDATAHEADER);
	rgn.rdh.iType    = RDH_RECTANGLES;       // Must be this
	rgn.rdh.nCount   = 2;                    // One rectangle
	rgn.rdh.nRgnSize = 32;                   // RECT Type is 16 bytes

	rgn.rdh.rcBound.left   = 0;
	rgn.rdh.rcBound.top    = 0;
	rgn.rdh.rcBound.right  = WINDOW_WIDTH - 1;
	rgn.rdh.rcBound.bottom = WINDOW_HEIGHT - 1;

	rgn.rc[0].left   = 0;
	rgn.rc[0].top    = 0;
	rgn.rc[0].right  = WINDOW_WIDTH / 2 - 1;
	rgn.rc[0].bottom = WINDOW_HEIGHT / 2 - 1;

	rgn.rc[1].left   = WINDOW_WIDTH / 2;
	rgn.rc[1].top    = WINDOW_HEIGHT / 2;
	rgn.rc[1].right  = WINDOW_WIDTH - 1;
	rgn.rc[1].bottom = WINDOW_HEIGHT - 1;

	// Create a clipper
	ddRet = IDirectDraw7_CreateClipper(lpDD, 0, &lpDDClipBack, NULL);
	if (ddRet != DD_OK){
		DDInitError("DDCreateClipper Failed", ddRet);
		return 1;
	}

	ddRet = IDirectDrawClipper_SetClipList(lpDDClipBack, (RGNDATA*)&rgn, 0);

	// Create a clipper for the window
	ddRet = IDirectDraw7_CreateClipper(lpDD, 0, &lpDDClipPrimary, NULL);
	if (ddRet != DD_OK){
		DDInitError("DDCreateClipper Failed", ddRet);
		return 1;
	}

	ddRet = IDirectDrawClipper_SetHWnd(lpDDClipPrimary, 0, hwndMain);
	if (ddRet != DD_OK){
		DDInitError("DDClipper_SetHWnd Failed", ddRet);
		return 1;
	}

	ddRet = ClearFrame(lpDDSBack, backCol);

	// attach the clipper to the front surface
	ddRet = IDirectDrawSurface7_SetClipper(lpDDSPrimary, lpDDClipPrimary);
	ddRet = IDirectDrawSurface7_SetClipper(lpDDSBack, lpDDClipBack);
	if (ddRet != DD_OK){
		DDInitError("DDSurface_SetClipper Failed", ddRet);
		return 1;
	}

	// Load small bitmap and make another surface
	ddRet = DDXCreateSurfaceFromBitmap(lpDD, fName, &lpBmpSurface);
	if( ddRet != DD_OK ){
		DDInitError("DDXCreateSurfaceFromBitmap Failed", ddRet);
		return 1;
	}

	// Set the color key for this image (black for this example) for sprites Transparent blit.
	DDXSetColorKey(lpBmpSurface, RGB(0, 0, 0));
	return DD_OK;
}

int UpdateFrame(HWND hwnd)
{
	static int count = 0, iPastCount;
	static char sText[80];
	int dCount, iCurrCount, iTickDiff, ddret;
	HDC hdc;

	ddret = ClearFrame(lpDDSBack, backCol);

	// Display the frames per second
	// This first time will be behind the sprites
	newTime = GetTickCount();
	iCurrCount = newTime;
	count++;

	if (iCurrCount - iPastCount > 1000){
		sprintf(sText, "Fms/Sec %d", count);
		count = 0;
		iPastCount = iCurrCount;
	}

	iTickDiff = newTime - oldTime;

	oldTime = newTime;

	ddret = IDirectDrawSurface7_GetDC(lpDDSBack, &hdc);
	if (ddret == DD_OK){
		SetTextColor(hdc, RGB(255, 255, 255));
		SetBkMode(hdc, TRANSPARENT);
		TextOut(hdc, 20, 20, sText, strlen(sText));
		IDirectDrawSurface7_ReleaseDC(lpDDSBack, hdc);
	}

	// Check the cooperative level before rendering
	ddret = IDirectDraw7_TestCooperativeLevel(lpDD);
	if (ddret != DD_OK){
		switch (ddret){

		case DDERR_EXCLUSIVEMODEALREADYSET:
			// Do nothing because some other app has exclusive mode
			sleep(10);
			return DD_OK;

		case DDERR_WRONGMODE:
			// The display mode changed on us. Update the DirectDraw surfaces
			// accordingly, destroy and remake.
			DestroyDDObjects();
			SetForegroundWindow(hwndMain);
			return CreateWindowedDisplay(hwndMain, WINDOW_WIDTH, WINDOW_HEIGHT);

		}
		return ddret;
	}

	for(int i = 0; i<NUMSPRITES; i++){
		// Move the sprites according to how much time has passed
		UpdateSprite(&Sprite[i], ((float)iTickDiff / 100.0f));
		ddret = IDirectDrawSurface7_Blt(lpDDSBack, (RECT*)&Sprite[i], lpBmpSurface, NULL, DDBLT_WAIT | DDBLT_KEYSRC, NULL);
		if (ddret == DDERR_SURFACELOST)
			ddret = RestoreAll();
	}

	// Display the frames per second
	// This second time will be in front the sprites
	ddret = IDirectDrawSurface7_GetDC(lpDDSBack, &hdc);
	if (ddret == DD_OK){
		SetTextColor(hdc, RGB(255, 255, 255));
		SetBkMode(hdc, TRANSPARENT);
		TextOut(hdc, 160, 160, sText, strlen(sText));
		IDirectDrawSurface7_ReleaseDC(lpDDSBack, hdc);
	}

	return FlipWindowed(hwnd, lpDDSPrimary, lpDDSBack);
}

static BOOL InitApplication(void)
{
	WNDCLASS wc;

	memset(&wc,0,sizeof(WNDCLASS));
	wc.style 			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc 		= (WNDPROC)MainWndProc;
	wc.hInstance 		= hInst;
	wc.hbrBackground 	= (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszClassName 	= "ddrawWndClass";
	wc.lpszMenuName 	= MAKEINTRESOURCE(IDMAINMENU);
	wc.hCursor 			= LoadCursor(NULL,IDC_ARROW);
	wc.hIcon 			= LoadIcon(NULL,IDI_APPLICATION);
	if (!RegisterClass(&wc))
		return 0;
	return 1;
}

HWND CreateddrawWndClassWnd(void)
{
	// The backsurface size will be WINDOW_WIDTH x WINDOW_HEIGHT
	// so adjust the main window so that the client dimensions
	// are the same. The reason is that Surface_Blt() stretches,
	// either shrinking or expanding if the two are not the same.
	int dwFrameWidth    = GetSystemMetrics(SM_CXSIZEFRAME);
	int dwFrameHeight   = GetSystemMetrics(SM_CYSIZEFRAME);
	int dwMenuHeight    = GetSystemMetrics(SM_CYMENU);
	int dwCaptionHeight = GetSystemMetrics(SM_CYCAPTION);
	int dwWindowWidth   = WINDOW_WIDTH  + dwFrameWidth * 2;
	int dwWindowHeight  = WINDOW_HEIGHT + dwFrameHeight * 2 + dwMenuHeight + dwCaptionHeight;

	return CreateWindow("ddrawWndClass","DDraw Clip",
		WS_MINIMIZEBOX | WS_VISIBLE | WS_CAPTION | WS_BORDER | WS_SYSMENU | WS_THICKFRAME,
		CW_USEDEFAULT,dwWindowWidth,CW_USEDEFAULT,dwWindowHeight,
		NULL,
		NULL,
		hInst,
		NULL);
}

LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch (msg) {
	case WM_ACTIVATE:
		bActive = LOWORD(wParam);
		return 0;

	case WM_SIZE:
		// Check to see if we are losing our window...
		if (SIZE_MAXHIDE == wParam || SIZE_MINIMIZED == wParam)
			bActive = FALSE;
		else if (SIZE_RESTORED == wParam)
			bActive = TRUE;
 		return 0;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDM_EXIT){
    		PostMessage(hwnd, WM_CLOSE, 0, 0);

		}else if (LOWORD(wParam) == IDM_CLIP){
			if( MF_CHECKED == GetMenuState(GetMenu(hwnd), IDM_CLIP, MF_BYCOMMAND)){
				rgn.rdh.nCount   = 1;                   // Make it one rectangle
				rgn.rc[0].left   = 0;
				rgn.rc[0].top    = 0;
				rgn.rc[0].right  = WINDOW_WIDTH - 1;
				rgn.rc[0].bottom = WINDOW_HEIGHT - 1;
				IDirectDrawClipper_SetClipList(lpDDClipBack, (RGNDATA*)&rgn, 0);
				// UnSet menu check mark
				CheckMenuItem(GetSubMenu(GetMenu(hwnd), 0), IDM_CLIP, MF_BYCOMMAND | MF_UNCHECKED);
			}else{
				rgn.rdh.nCount   = 2;                     // For two rectangles.
				rgn.rc[0].left   = 0;                     //
				rgn.rc[0].top    = 0;                     // Only have to re-write the
				rgn.rc[0].right  = WINDOW_WIDTH / 2 - 1;  // first one because the other
				rgn.rc[0].bottom = WINDOW_HEIGHT / 2 - 1; // has not been changed

				// Set menu check mark
				CheckMenuItem(GetSubMenu(GetMenu(hwnd), 0), IDM_CLIP, MF_BYCOMMAND | MF_CHECKED);
				ClearFrame(lpDDSBack, backCol);
				IDirectDrawClipper_SetClipList(lpDDClipBack, (RGNDATA*)&rgn, 0);
			}

		}
		return 0;

	case WM_GETMINMAXINFO:{
	    // Don't allow resizing.
    	// Fix the size of the window to (client size)
    	MINMAXINFO * pMinMax = (MINMAXINFO *)lParam;

    	int dwFrameWidth    = GetSystemMetrics(SM_CXSIZEFRAME);
    	int dwFrameHeight   = GetSystemMetrics(SM_CYSIZEFRAME);
    	int dwMenuHeight    = GetSystemMetrics(SM_CYMENU);
    	int dwCaptionHeight  = GetSystemMetrics(SM_CYCAPTION);

    	pMinMax->ptMinTrackSize.x = WINDOW_WIDTH  + dwFrameWidth * 2;
    	pMinMax->ptMinTrackSize.y = WINDOW_HEIGHT + dwFrameHeight * 2 + dwMenuHeight + dwCaptionHeight;

    	pMinMax->ptMaxTrackSize.x = pMinMax->ptMinTrackSize.x;
    	pMinMax->ptMaxTrackSize.y = pMinMax->ptMinTrackSize.y;
		}
		return 0;

	case WM_EXITMENULOOP:
	    // Ignore time spent in menu
		oldTime = GetTickCount();
	    return 0;

	case WM_EXITSIZEMOVE:
	    // Ignore time spent resizing or moving
		oldTime = GetTickCount();
		return 0;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;

	default:
		return DefWindowProc(hwnd, msg, wParam, lParam);
	}
}

void DestroyDDObjects(void)
{
	// After Releasing Objects set Objects to NULL
	if (lpDD != NULL){

		if (lpDDSPrimary != NULL){
			lpDDSPrimary->lpVtbl->Release(lpDDSPrimary);
			lpDDSPrimary = NULL;
		}

		if (lpDDSBack != NULL){
			lpDDSBack->lpVtbl->Release(lpDDSBack) ;
			lpDDSBack = NULL;
		}

		if (lpDDClipBack != NULL){
			lpDDClipBack->lpVtbl->Release(lpDDClipBack) ;
			lpDDClipBack = NULL;
		}

		if (lpDDClipPrimary != NULL){
			lpDDClipPrimary->lpVtbl->Release(lpDDClipPrimary) ;
			lpDDClipPrimary = NULL;
		}

		if (lpBmpSurface != NULL){
			lpBmpSurface->lpVtbl->Release(lpBmpSurface) ;
			lpBmpSurface = NULL;
		}

		lpDD->lpVtbl->Release(lpDD) ;
		lpDD = NULL;
	}
}

int ClearFrame(LPDIRECTDRAWSURFACE7 pBackBuffer, DWORD dwColor)
{
	int ddret;

	if (NULL == pBackBuffer)
    	return E_POINTER;

	// Erase the background
	DDBLTFX ddbltfx;
	memset(&ddbltfx, 0, sizeof(DDBLTFX));
	ddbltfx.dwSize      = sizeof(DDBLTFX);
	ddbltfx.dwFillColor = dwColor;

	ddret = IDirectDrawSurface7_Blt(pBackBuffer, 0, 0, 0, DDBLT_COLORFILL, &ddbltfx);

	if (ddret == DDERR_SURFACELOST)
		ddret = RestoreAll();
	else if (ddret != DD_OK)
    	return ddret;

	return ddret;

}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, char * lpCmdLine, int nCmdShow)
{
	MSG msg;
	HANDLE hAccelTable;

	hInst = hInstance;
	if (!InitApplication())
		return 0;

	if ((hwndMain = CreateddrawWndClassWnd()) == (HWND)0)
		return 0;

	// Set the sprite's position and velocity
	for (int i = 0; i<NUMSPRITES; i++){
		Sprite[i].X  = Random(WINDOW_WIDTH);
		Sprite[i].fX = (float)Sprite[i].X;
		Sprite[i].Y  = Random(WINDOW_HEIGHT);
		Sprite[i].fY = (float)Sprite[i].Y;
		Sprite[i].W  = SPRITE_WIDTH;
		Sprite[i].H  = SPRITE_HEIGHT;

		if (i & 1){ // odd
			Sprite[i].dX = Random(15) + 10;
			Sprite[i].dY = -Random(15) - 10;
		}else{		// even
			Sprite[i].dX = -Random(15) - 10;
			Sprite[i].dY = Random(15) + 10;
		}
	}

	// CreateWindowedDisplay does the job of initialising DDraw
	// and creating the windowed display
	if(CreateWindowedDisplay(hwndMain, WINDOW_WIDTH, WINDOW_HEIGHT))
		goto Cleanup;

	ShowWindow(hwndMain, SW_SHOW);

	// Set menu check mark
	CheckMenuItem(GetSubMenu(GetMenu(hwndMain), 0), IDM_CLIP, MF_BYCOMMAND | MF_CHECKED);

	oldTime = GetTickCount();

    while( 1 ) {
        if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) ){
            if( !GetMessage( &msg, NULL, 0, 0 ) ) {
                break;
            }
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }else if( bActive ){
			UpdateFrame(hwndMain);
        }else{
            WaitMessage();			 // Wait for message to re-activate app.
			oldTime = GetTickCount();// Ignore time spent inactive
        }
    }

Cleanup:
	DestroyDDObjects();

	return 0;
}
